# Copyright (c) 2012, 2015, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA


from wb import DefineModule
import grt
import mforms

from workbench.log import log_warning, log_error
from workbench.utils import Version
from workbench.graphics.cairo_utils import Context
from explain_renderer import ExplainContext, decode_json

ModuleInfo = DefineModule(name= "SQLIDEQueryAnalysis", author= "Oracle Corp.", version= "1.0")


class JSONTreeViewer(mforms.TreeNodeView):
    def __init__(self):
        mforms.TreeNodeView.__init__(self, mforms.TreeAltRowColors|mforms.TreeShowColumnLines|mforms.TreeShowRowLines)
        self.add_column(mforms.StringColumnType, "Key", 200)
        self.add_column(mforms.StringColumnType, "Value", 300)
        self.end_columns()

    def display_data(self, data):
        def add_nodes(node, create_node, data):
            if type(data) is dict:
                if node:
                    node.set_string(1, "<dict>")
                for key, value in data.items():
                    ch = create_node()
                    ch.set_string(0, key)
                    add_nodes(ch, ch.add_child, value)
            elif type(data) is list:
                if node:
                    node.set_string(1, "<list>")
                for i, value in enumerate(data):
                    ch = create_node()
                    ch.set_string(0, str(i))
                    add_nodes(ch, ch.add_child, value)
            else:
                if not node:
                    node = create_node()
                if type(data) is bool:
                    node.set_string(1, "true" if data else "false")
                else:
                    node.set_string(1, str(data))
    
        data = decode_json(data)
        self.clear()
        add_nodes(None, self.add_node, data)


class RenderBox(mforms.PyDrawBox):
    def __init__(self, context, scroll):
        mforms.PyDrawBox.__init__(self)
        self.set_managed()
        self.set_release_on_add()

        self.scroll = scroll

        self.set_instance(self)
        self.offset = (0, 0)
        self.size = None
        self.vertical = False
        self.node_spacing = 20
        self.econtext = context
    
        self.drag_offset = None
    
    def mouse_down(self, b, x, y):
        if b == 0:
            self.drag_offset = (x, y)
        x -= self.offset[0]
        y -= self.offset[1]
        self.econtext._canvas.mouse_down(b, x, y)
        if self.econtext.overview_mode and b == 0:
            self.econtext.mouse_down(x, y)
        self.econtext.close_tooltip()


    def mouse_up(self, b, x, y):
        if b == 0:
            self.drag_offset = None
        x -= self.offset[0]
        y -= self.offset[1]
        self.econtext._canvas.mouse_up(b, x, y)

    def mouse_move(self, x, y):
        x -= self.offset[0]
        y -= self.offset[1]
        # for drag panning
        #if self.drag_offset:
        #  self.scroll.scroll_to(self.drag_offset[0] - x, self.drag_offset[1] - y)
        self.econtext._canvas.mouse_move(x, y)
        if self.econtext.overview_mode:
            self.econtext.mouse_moved(x, y)
            self.set_needs_repaint()

    def mouse_leave(self):
        self.econtext._canvas.mouse_leave()

    def relayout(self):
        w, h = self.econtext.layout()
        if self.get_width() != w or self.get_height() != h:
            self.set_size(w, h)

    
    def repaint(self, cr, x, y, w, h):
        c = Context(cr)

        c.set_source_rgb(1, 0xfc/255.0, 0xe5/255.0)

        try:
            dw, dh = self.econtext.size
            xx = 0
            yy = 0
            if dw < self.get_width():
                xx = (self.get_width() - dw)/2
            if dh < self.get_height():
                yy = (self.get_height() - dh)/2
            self.offset = (xx, yy)
            self.econtext.set_offset(xx, yy)
            self.econtext.repaint(c)
        except Exception:
            import traceback
            log_error("Exception rendering explain output: %s\n" % traceback.format_exc())



def newToolBarItem(*args):
    item = mforms.ToolBarItem(*args)
    item.set_managed()
    return item


EXPLAIN_COLUMN_WIDTHS = {
  "id" : 40,
  "select_type" : 120,
  "table" : 100,
  "type" : 50,
  "possible_keys" : 200,
  "key" : 200,
  "key_len" : 40,
  "ref" : 200,
  "rows" : 50,
  "Extra" : 200
}


class QueryPlanTab(mforms.Box):
    node_spacing = 30
    vertical = True

    def __init__(self, owner, json_text, context, server_version):
        mforms.Box.__init__(self, False)
        self.set_managed()
        self.set_release_on_add()

        self._context = context

        get_resource_path = mforms.App.get().get_resource_path

        self.toolbar = mforms.newToolBar(mforms.SecondaryToolBar)
        self.toolbar.set_back_color("#ffffff")

        self.switcher_item = newToolBarItem(mforms.SelectorItem)
        self.toolbar.add_item(self.switcher_item)

        s = newToolBarItem(mforms.SeparatorItem)
        self.toolbar.add_item(s)

        l = newToolBarItem(mforms.LabelItem)
        l.set_text("Display Info:")
        self.toolbar.add_item(l)

        item = newToolBarItem(mforms.SelectorItem)
        item.set_selector_items(["Read + Eval cost", "Data Read per Join"])
        item.add_activated_callback(self.display_cost)
        self.toolbar.add_item(item)
        cost_type_item = item

        # cost info was added in 5.7.2
        has_cost_info = server_version >= Version(5, 7)
        if not has_cost_info:
            item.set_enabled(False)

#item = newToolBarItem(mforms.SelectorItem)
#        item.set_selector_items(["Show Aggregated Costs", "Show Individual Costs"])
#        item.add_activated_callback(self.toggle_aggregated)
#        self.toolbar.add_item(item)


        #btn = newToolBarItem(mforms.SegmentedToggleItem)
        #btn.set_icon(get_resource_path("qe_resultset-tb-switcher_grid_off_mac.png"))
        #btn.set_alt_icon(get_resource_path("qe_resultset-tb-switcher_grid_on_mac.png"))
        #self.toolbar.add_item(btn)

        #btn = newToolBarItem(mforms.SegmentedToggleItem)
        #btn.set_icon(get_resource_path("qe_resultset-tb-switcher_explain_off.png"))
        #btn.set_alt_icon(get_resource_path("qe_resultset-tb-switcher_explain_on.png"))
        #self.toolbar.add_item(btn)

        s = newToolBarItem(mforms.SeparatorItem)
        self.toolbar.add_item(s)

        btn = newToolBarItem(mforms.ActionItem)
        btn.set_icon(get_resource_path("tiny_saveas.png"))
        btn.add_activated_callback(self.save)
        btn.set_tooltip("Save image to an external file.")
        self.toolbar.add_item(btn)

        s = newToolBarItem(mforms.SeparatorItem)
        self.toolbar.add_item(s)

        l = newToolBarItem(mforms.LabelItem)
        l.set_text("Overview:")
        self.toolbar.add_item(l)

        btn = newToolBarItem(mforms.ActionItem)
        btn.set_icon(get_resource_path("qe_sql-editor-explain-tb-overview.png"))
        btn.add_activated_callback(self.overview)
        btn.set_tooltip("Zoom out the diagram.")
        self.toolbar.add_item(btn)

        s = newToolBarItem(mforms.SeparatorItem)
        self.toolbar.add_item(s)

        l = newToolBarItem(mforms.LabelItem)
        l.set_text("View Source:")
        self.toolbar.add_item(l)

        btn = newToolBarItem(mforms.ToggleItem)
        btn.set_icon(get_resource_path("qe_sql-editor-tb-icon_word-wrap-off.png"))
        btn.set_alt_icon(get_resource_path("qe_sql-editor-tb-icon_word-wrap-on.png"))
        btn.add_activated_callback(self.switch_to_raw)
        btn.set_tooltip("View the raw JSON explain data.")
        self.toolbar.add_item(btn)


        self.add(self.toolbar, False, True)

        # Query Plan diagram
        self.scroll = mforms.newScrollPanel(mforms.ScrollPanelNoFlags)
        self.scroll.set_back_color("#ffffff")
        self.scroll.set_visible_scrollers(True, True)
        
        #self.img = mforms.newImageBox()
        self.drawbox = RenderBox(self._context, self.scroll)
        self.scroll.add(self.drawbox)
        
        self.drawbox.node_spacing = self.node_spacing
        self.drawbox.vertical = self.vertical
        self.add(self.scroll, True, True)
    
        self.display_cost(cost_type_item)

        # textbox to view the json data
        self._raw_explain = mforms.CodeEditor()
        self._raw_explain.set_value(json_text)
        #self._raw_explain.enable_folding(True)
        self._raw_explain.set_language(mforms.LanguagePython)
        self._raw_explain.set_features(mforms.FeatureReadOnly, True)
        self.add(self._raw_explain, True, True)
        self._raw_explain.show(False)


    def display_cost(self, item):
        text = item.get_text()
        if text:
            cost = text.lower().split()[0]
            if cost == "read":
                self._context.show_cost_info_type("read_eval_cost")
            elif cost == "data":
                self._context.show_cost_info_type("data_read_per_join")
            else:
                grt.log_error("vexplain", "Unknown cost info type: %s\n" % cost)
            self.drawbox.set_needs_repaint()


    def toggle_aggregated(self, item):
        self._context.show_aggregated_cost_info("aggregated" in item.get_text().lower())
        self.drawbox.set_needs_repaint()


    def switch_to_raw(self, item):
        flag = item.get_checked()
        self._raw_explain.show(flag)
        self.scroll.show(not flag)


    def save(self, item):
        ch = mforms.FileChooser(mforms.SaveFile)
        directory = grt.root.wb.options.options.get("wb.VisualExplain:LastFileChooserDirectory", "")
        if directory:
            ch.set_directory(directory)
        ch.set_extensions("PNG image (*.png)|*.png", "png")
        ch.set_title("Save Image As")
        ch.set_path("explain.png")
        if ch.run_modal():
            self._context.export_to_png(ch.get_path())
            grt.root.wb.options.options["wb.VisualExplain:LastFileChooserDirectory"] = ch.get_directory()


    def set_needs_repaint(self, x, y, w, h):
        self.drawbox.set_needs_repaint()


    def overview(self, item):
        self._context.enter_overview_mode()
        self.drawbox.set_needs_repaint()



class TabularExplainTab(mforms.Box):
    node_spacing = 30
    vertical = True

    def __init__(self, owner, explain, server_version):
        mforms.Box.__init__(self, False)
        self.set_managed()
        self.set_release_on_add()

        self.toolbar = mforms.newToolBar(mforms.SecondaryToolBar)
        self.toolbar.set_back_color("#ffffff")

        self.switcher_item = newToolBarItem(mforms.SelectorItem)
        self.toolbar.add_item(self.switcher_item)

        self.add(self.toolbar, False, False)

        self.explain_tree = mforms.newTreeNodeView(mforms.TreeFlatList|mforms.TreeShowColumnLines|mforms.TreeShowRowLines|mforms.TreeAltRowColors)
        self.explain_tree.add_column_resized_callback(self.column_resized)
        c = len(explain.columns)
        rows_column = None

        saved_widths = grt.root.wb.state.get("wb.query.analysis:ExplainTreeColumnWidths", None)
        if saved_widths:
          saved_widths = [int(i) for i in saved_widths.split(",")]

        for i, column in enumerate(explain.columns):
          width = saved_widths[i] if saved_widths and i < len(saved_widths) else EXPLAIN_COLUMN_WIDTHS.get(column.name, 100)
          if column.name == "rows":
            rows_column = i
            self.explain_tree.add_column(mforms.LongIntegerColumnType, column.name, width)
          else:
            self.explain_tree.add_column(mforms.StringColumnType, column.name, width)
        self.explain_tree.end_columns()

        if explain.goToFirstRow():
          while True:
            node = self.explain_tree.add_node()
            for i in range(c):
              value = explain.stringFieldValue(i)
              if i == rows_column:
                node.set_long(i, long(value) if value else 0)
              else:
                node.set_string(i, value)
            if not explain.nextRow():
              break

        explain.reset_references()
        self.add(self.explain_tree, True, True)


    def column_resized(self, column):
        sizes = []
        for i in range(self.explain_tree.get_column_count()):
            sizes.append(self.explain_tree.get_column_width(i))
        grt.root.wb.state["wb.query.analysis:ExplainTreeColumnWidths"] = ",".join([str(i) for i in sizes])


class ExplainTab(mforms.AppView):
    _query_plan = None
    _costs_tree = None
    _explain_context = None

    def __init__(self, server_version, query, json_text, explain):
        mforms.AppView.__init__(self, False, "QueryExplain", False)
        self.set_managed()
        self.set_release_on_add()

        self.on_close(self.on_tab_close)

        self._form_deactivated_conn = mforms.Form.main_form().add_deactivated_callback(self.form_deactivated)

        if not json_text and not explain:
            label = mforms.newLabel("Explain data not available for statement")
            label.set_style(mforms.BigBoldStyle)
            label.set_text_align(mforms.MiddleCenter)
            self.add(label, True, True)
            return

        self._query = query
        self.tabview = mforms.newTabView(mforms.TabViewTabless)
        self.tabview.add_tab_changed_callback(self.tab_changed)
        self.set_back_color("#ffffff")
        
        default_tab = grt.root.wb.state.get("wb.query.analysis:ActiveExplainTab", 0)
        json_data = None
        if json_text:
            try:
                json_data = decode_json(json_text)
            except Exception, e:
                import traceback
                log_error("Error creating query plan: %s\n" % traceback.format_exc())
                mforms.Utilities.show_error("Query Plan Generation Error",
                                            "An unexpected error occurred parsing JSON query explain data.\nPlease file a bug report at http://bugs.mysql.com along with the query and the Raw Explain Data.\n\nException: %s" % e,
                                            "OK", "", "")
        if json_data:
            try:
                self._explain_context = ExplainContext(json_data, server_version)

                self._query_plan = QueryPlanTab(self, json_text, self._explain_context, server_version)
                self._explain_context.init_canvas(self._query_plan.drawbox, self._query_plan.scroll, self._query_plan.set_needs_repaint)
                #self._explain_context._canvas.set_background_color(1, 0xfc/255.0, 0xe5/255.0)
                # Initial layouting of the plan diagram
                self._query_plan.drawbox.relayout()
                self.tabview.add_page(self._query_plan, "Visual Explain")

                self._query_plan.switcher_item.set_name("visual_explain_switcher")
                self._query_plan.switcher_item.set_selector_items(["Visual Explain", "Tabular Explain"])
                self._query_plan.switcher_item.add_activated_callback(self.switch_view)
            except Exception, e:
                import traceback
                log_error("Error creating query plan: %s\n" % traceback.format_exc())
                mforms.Utilities.show_error("Query Plan Generation Error",
                                            "An unexpected error occurred during creation of the graphical query plan.\nPlease file a bug report at http://bugs.mysql.com along with the query and the Raw Explain Data.\n\nException: %s" % e,
                                            "OK", "", "")
        else:
            log_error("No JSON data for explain\n")

        # Good old explain
        if explain:
            self._tabular_explain = TabularExplainTab(self, explain, server_version)
            self.tabview.add_page(self._tabular_explain, "Tabular Explain")

            self._tabular_explain.switcher_item.set_name("tabular_explain_switcher")
            self._tabular_explain.switcher_item.set_selector_items(["Visual Explain", "Tabular Explain"])
            self._tabular_explain.switcher_item.set_text("Tabular Explain")
            self._tabular_explain.switcher_item.add_activated_callback(self.switch_view)


        self.add(self.tabview, True, True)

        if self._explain_context:
            self.tabview.set_active_tab(default_tab)


    _switching = False
    def switch_view(self, item):
        if self._switching: return
        self._switching = True
        new_view = item.get_text()
        if new_view == "Visual Explain":
            self.tabview.set_active_tab(0)
        elif new_view == "Tabular Explain":
            self.tabview.set_active_tab(1)
        else:
            raise Exception("Unknown "+new_view)

        source = item.get_name()
        # switch back the selector
        if source == "visual_explain_switcher":
          item.set_text("Visual Explain")
        elif source == "tabular_explain_switcher":
          item.set_text("Tabular Explain")
        self._switching = False


    def tab_changed(self):
        if self._query_plan and self._explain_context:
            if self.tabview.get_active_tab() == 0:
                self._explain_context._canvas.activate()
            else:
                self._explain_context._canvas.deactivate()
                self._explain_context.close_tooltip()
            grt.root.wb.state["wb.query.analysis:ActiveExplainTab"] = self.tabview.get_active_tab()


    def fill_costs_tree(self, json):
        tree = mforms.newTreeNodeView(mforms.TreeFlatList|mforms.TreeShowColumnLines|mforms.TreeShowRowLines|mforms.TreeAltRowColors)
        return tree


    def form_deactivated(self):
        if self._query_plan and self._explain_context:
            self._explain_context.close_tooltip()

    def on_tab_close(self):
        self.form_deactivated()
        if self._form_deactivated_conn:
            self._form_deactivated_conn.disconnect()
            self._form_deactivated_conn = None
        return True


@ModuleInfo.export(grt.INT, grt.classes.db_query_QueryEditor, grt.classes.db_query_ResultPanel)
def visualExplain(editor, result_panel):
    version = Version.fromgrt(editor.owner.serverVersion)
    
    statement = editor.currentStatement
    if statement:
        try:
            explain = editor.owner.executeQuery("EXPLAIN %s" % statement, 1)
        except Exception, e:
            log_warning("Could not execute EXPLAIN %s: %s\n" % (statement, e))
            explain = None

        json = None
        if explain and version.is_supported_mysql_version_at_least(5, 6):
            rset = editor.owner.executeQuery("EXPLAIN FORMAT=JSON %s" % statement, 1)
            if rset and rset.goToFirstRow():
                json = rset.stringFieldValue(0)
                rset.reset_references()

        view = ExplainTab(version, statement, json, explain if explain else None)
        dock = mforms.fromgrt(result_panel.dockingPoint)
        view.set_identifier("execution_plan")
        view.set_title("Execution\nPlan")
        dock.dock_view(view, "output_type-executionplan.png", 0)
        dock.select_view(view)
    
    return 0


@ModuleInfo.export(grt.INT, grt.classes.db_query_QueryEditor, grt.STRING, grt.STRING)
def visualExplainForConnection(editor, conn_id, the_query):
    version = Version.fromgrt(editor.owner.serverVersion)

    try:
        explain = editor.owner.executeManagementQuery("EXPLAIN FOR CONNECTION %s" % conn_id, 1)
    except grt.DBError, e:
        if e.args[1] == 0:
            mforms.Utilities.show_message("Explain for Connection",
                                          "Explain for connection %s did not generate any output." % conn_id, "OK", "", "")
        else:
            mforms.Utilities.show_error("Explain for Connection",
                                        "Error executing explain for connection %s\n%s" % (conn_id, e), "OK", "", "")
        return 0
    except Exception, e:
        mforms.Utilities.show_error("Explain for Connection",
                                    "Error executing explain for connection %s\n%s" % (conn_id, e), "OK", "", "")
        return 0

    if not explain:
        mforms.Utilities.show_error("Explain for Connection",
                                    "Error executing explain for connection %s" % conn_id, "OK", "", "")
        return 0
    
    if version.is_supported_mysql_version_at_least(5, 6):
        rset = editor.owner.executeManagementQuery("EXPLAIN FORMAT=JSON FOR CONNECTION %s" % conn_id, 1)
        if rset and rset.goToFirstRow():
            json = rset.stringFieldValue(0)
            rset.reset_references()
        else:
            json = None
    
    view = ExplainTab(version, the_query, json, explain if explain else None)
    view.set_identifier("execution_plan")
    dock = mforms.fromgrt(editor.resultDockingPoint)
    dock.dock_view(view, "", 0)
    dock.select_view(view)
    view.set_title("Explain for Connection")
    
    return 0




# Copyright (c) 2012, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import functools

import grt
import mforms

from grt import DBLoginError
from workbench.ui import WizardPage

from migration_source_selection import request_password, test_connectivity


class SourceTargetMainView(WizardPage):
    def __init__(self, main):
        super(SourceTargetMainView, self).__init__(main, 'Connection Selection')
        self._connections = []
        
    def create_ui(self):
        self.content.set_padding(20)

        self.content.add(mforms.newLabel('''Select the connection for the source MySQL server where databases will be copied from
and the destination server where they should be copied to.'''), False, True)

        source_panel = mforms.newPanel(mforms.TitledBoxPanel)
        source_panel.set_title('Source MySQL Server')

        box = mforms.newBox(False)
        box.set_spacing(12)
        box.set_padding(8)
        box.add(mforms.newLabel('Select the connection for the source MySQL server instance'), False, True)
        
        source_hbox = mforms.newBox(True)
        source_hbox.set_spacing(8)
        source_hbox.add(mforms.newLabel('Source Connection:'), False, True)
        self.source_selector = mforms.newSelector()
        source_hbox.add(self.source_selector, True, True)
        box.add(source_hbox, False, True)

        self.source_connection_status = mforms.newLabel('')
        self.source_connection_status.set_color('#aa3333')
        box.add(self.source_connection_status, True, True)
        self.source_selector.add_changed_callback(functools.partial(self.selector_changed, label=self.source_connection_status, selector=self.source_selector))

        source_panel.add(box)
        self.content.add(source_panel, False, True)

        target_panel = mforms.newPanel(mforms.TitledBoxPanel)
        target_panel.set_title('Destination MySQL Server')

        tbox = mforms.newBox(False)
        tbox.set_spacing(12)
        tbox.set_padding(8)
        tbox.add(mforms.newLabel('Select the connection object for the destination MySQL server instance'), False, True)
        
        target_hbox = mforms.newBox(True)
        target_hbox.set_spacing(8)
        target_hbox.add(mforms.newLabel('Target Connection:'), False, True)
        self.target_selector = mforms.newSelector()
        target_hbox.add(self.target_selector, True, True)
        tbox.add(target_hbox, False, True)

        self.target_connection_status = mforms.newLabel('')
        self.target_connection_status.set_color('#aa3333')
        tbox.add(self.target_connection_status, True, True)
        self.target_selector.add_changed_callback(functools.partial(self.selector_changed, label=self.target_connection_status, selector=self.target_selector))
        
        target_panel.add(tbox)
        self.content.add(target_panel, False, True)

        self.advanced_button.set_text('Test Connections')

        self.back_button.set_enabled(False)

        self.load_connections()
    
    def selector_changed(self, label, selector):
        self.connections_ok = False
        label.set_text('')

        if selector.get_selected_index() == selector.get_item_count()-1:
            grt.modules.Workbench.showConnectionManager()
            self.load_connections()

    def load_connections(self):
        def formatConnection(conn):
            i = conn.hostIdentifier
            return i[i.find('@')+1:]

        # filter out ssh connections until its supported
        self._connections = [conn for conn in grt.root.wb.rdbmsMgmt.storedConns if conn.driver and not conn.driver.name.endswith("SSH")]
        selector_items = ( ['Pick a Connection'] +
            ['%s (%s)' % (conn.name, formatConnection(conn)) for conn in self._connections] +
            ['-', 'Edit Connections...'] )

        self.source_selector.clear()
        self.source_selector.add_items(selector_items)
        self.source_connection_status.set_text('')
        self.source_selector.set_selected(0)
        
        self.target_selector.clear()
        self.target_selector.add_items(selector_items)
        self.target_connection_status.set_text('')
        self.target_selector.set_selected(0)

        self.connections_ok = False


    def test_connection(self, source, caption):
        info_label = self.source_connection_status if caption=='Source' else self.target_connection_status
        info_label.set_text('Testing network connectivity...')
        info_label.set_color('#aa3333')
        if test_connectivity(source.connection, 'Test %s DBMS Connection' % caption) == False:
            info_label.set_text('Server could not be contacted')
            return
        info_label.set_text('Testing connection to DBMS...')
 
        force_password = False
        attempt = 0
        while True:
            try:
                if not source.connect():
                    info_label.set_text('Could not connect to DBMS')
                    self.connections_ok = False
                    return
                info_label.set_color('#33aa33')
                info_label.set_text('Connection succeeded.')
                self.connections_ok = True
                break
            except (DBLoginError, SystemError), e:
                if attempt > 0:
                    if isinstance(e, DBLoginError) and not force_password:
                        force_password = True
                    else:
                        etext = str(e)
                        if etext.startswith('Error(') and ': error calling ' in etext:
                            try:
                                etext = eval(etext[7:etext.rfind('):')-1], {}, {})[1]
                            except:
                                pass
                        info_label.set_text('Could not connect to DBMS: %s' % etext)
                        self.connections_ok = False
                        return
                attempt += 1
                source.password = request_password(source.connection, force_password)
            except Exception, e:
                etext = str(e)
                if etext.startswith('Error(') and etext.endswith(')'):
                    etext = eval(etext[6:-1], {}, {})[1]
                info_label.set_text('Could not connect to DBMS: %s' % etext)
                self.connections_ok = False


    @property
    def source_connection(self):
        i = self.source_selector.get_selected_index()
        return  self._connections[i-1] if 0 < i < len(self._connections)+1 else None


    @property
    def target_connection(self):
        i = self.target_selector.get_selected_index()
        return  self._connections[i-1] if 0 < i < len(self._connections)+1 else None


    def go_advanced(self):
        if self.source_connection:
            self.main.plan.setSourceConnection(self.source_connection)
            self.test_connection(self.main.plan.migrationSource, 'Source')
        else:
            self.source_connection_status.set_text('Please select a connection')
            self.source_connection_status.set_color('#aa3333')
            self.connections_ok = False

        if self.target_connection and self.source_connection == self.target_connection:
            self.target_connection_status.set_text('Select different connections for source and target')
            self.target_connection_status.set_color('#aa3333')
            self.connections_ok = False
        elif self.target_connection:
            self.main.plan.setTargetConnection(self.target_connection)
            self.test_connection(self.main.plan.migrationTarget, 'Target')
        else:
            self.target_connection_status.set_text('Please select a connection')
            self.target_connection_status.set_color('#aa3333')
            self.connections_ok = False


    def go_next(self):
        if not self.connections_ok:
            self.go_advanced()
            if not self.connections_ok:
                return

        set_status_text = mforms.App.get().set_status_text
        set_status_text('Connecting to source DBMS...')
        # Set source connection and connect:
        self.main.plan.setSourceConnection(self.source_connection)
        self.main.plan.migrationSource.connect()

        # Set target connection and connect:
        set_status_text('Connecting to target DBMS...')
        self.main.plan.setTargetConnection(self.target_connection)
        self.main.plan.migrationTarget.connect()

        # Fetch schema names:
        set_status_text('Fetching schema names...')
        self.main.plan.migrationSource.doFetchSchemaNames()
        
        set_status_text('Ready')
        super(SourceTargetMainView, self).go_next()



import pyodbc

import db_mssql_test_main

from workbench import db_driver

class TestPythonDBDriver(db_mssql_test_main.MssqlTestCase):

    @classmethod
    def setUpClass(cls):
        cls.conn = db_driver.connect(cls.connection, db_mssql_test_main.test_params['password'])

    def test_connection(self):
        self.assertIsInstance(self.conn, pyodbc.Connection)




# Copyright (c) 2012, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

from __future__ import with_statement
import platform
import threading
import random
import Queue
import traceback
import socket
import select
import sys
import time
import os

import mforms

import paramiko
from workbench.log import log_warning, log_error, log_debug, log_debug2, log_debug3, log_info
from wb_common import SSHFingerprintNewError, format_bad_host_exception

import grt

SSH_PORT = 22
REMOTE_PORT = 3306

# timeout for closing an unused tunnel
TUNNEL_TIMEOUT = 3

# paramiko 1.6 didn't have this class
if hasattr(paramiko, "WarningPolicy"):
    WarningPolicy = paramiko.WarningPolicy
else:
    class WarningPolicy(paramiko.MissingHostKeyPolicy):
        def missing_host_key(self, client, hostname, key):
            import binascii
            log_warning('WARNING: Unknown %s host key for %s: %s\n' % (key.get_name(), hostname, binascii.hexlify(key.get_fingerprint())))

class StoreIfConfirmedPolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        raise SSHFingerprintNewError("Key mismatched", client, hostname, key)

tunnel_serial=0
class Tunnel(threading.Thread):
    
    """This class is a threaded implementation of an SSH tunnel.
    
    You should not access the attributes that starts with an underscore outside this thread
    of execution (e.g. self._server) for this could run into race conditions. Even when
    accessing its public attributes (those that don't start with an underscore) you should
    be careful of acquiring the self.lock reentrant lock (and releasing it once done):
    
        with tunnel.lock:
            if tunnel.connecting:
                ... whatever...
                
    """

    def __init__(self, q, server, username, target, password, keyfile):
        super(Tunnel, self).__init__()
        self.daemon = True

        self._server = server
        self._username = username
        self._target = target
        self._password = password
        self._keyfile = keyfile
        
        # Acquire and release this lock while accessing the attributes of objects
        # of this class from the main thread: 
        self.lock = threading.RLock()
        
        # This event marks when a random port is selected to be used:
        self.port_is_set = threading.Event()

        self.local_port = None
        self._listen_sock = None
        self.q = q

        self._shutdown = False
        self.connecting = False
        
        self._client = paramiko.SSHClient()
        self._connections = []


    def is_connecting(self):
        with self.lock:
            return self.connecting

    def run(self):
        try:
            self.do_run()
        except Exception, e:
            log_error("Unhandled exception in SSH tunnel: %s\n" % e)


    def do_run(self):
        global tunnel_serial
        tunnel_serial += 1
        mforms.Utilities.set_thread_name("SSHTunnel%i"%tunnel_serial)

        sys.stdout.write('Thread started\n')  # sys.stdout.write is thread safe while print isn't
        log_debug2("SSH Tunel %i thread started\n" % tunnel_serial)
        # Create a socket and pick a random port number for it:
        self._listen_sock = socket.socket()
        while True:
                local_port = random.randint(1024, 65535)
                try:
                    self._listen_sock.bind(('127.0.0.1', local_port))
                    self._listen_sock.listen(2)
                    with self.lock:
                        self.local_port = local_port
                    break
                except socket.error, exc:
                    sys.stdout.write('Socket error: %s for port %d\n' % (exc, local_port) )
                    err, msg = exc.args
                    if err == 22:
                        continue # retry
                    self.notify_exception_error('ERROR',"Error initializing server end of tunnel", sys.exc_info())
                    raise exc
                finally:
                    with self.lock:
                        self.connecting = True

                    self.port_is_set.set()

        if self._keyfile:
            self.notify('INFO', 'Connecting to SSH server at %s:%s using key %s...' % (self._server[0], self._server[1], self._keyfile) )
        else:
            self.notify('INFO', 'Connecting to SSH server at %s:%s...' % (self._server[0], self._server[1]) )

        connected = self._connect_ssh()
        if not connected:
            self._listen_sock.close()
            self._shutdown = True

        with self.lock:
            self.connecting = False
            
        if connected:
            self.notify('INFO', 'Connection opened')

        del self._password
        last_activity = time.time()
        while not self._shutdown:
            try:
                socks = [self._listen_sock]
                for sock, chan in self._connections:
                    socks.append(sock)
                    socks.append(chan)
                r, w, x = select.select(socks, [], [], TUNNEL_TIMEOUT)
            except Exception, e:
                if not self._shutdown:
                    self.notify_exception_error('ERROR', 'Error while forwarding data: %r' % e, sys.exc_info())
                break

            if not r and len(socks) <= 1 and time.time() - last_activity > TUNNEL_TIMEOUT:
                self.notify('INFO', 'Closing tunnel to %s:%s for inactivity...' % (self._server[0], self._server[1]) )
                break
            last_activity = time.time()

            if self._listen_sock in r:
                self.notify('INFO', 'New client connection')
                self.accept_client()
            
            closed = []
            for sock, chan in self._connections:
                if sock in r:
                    data = sock.recv(1024)
                    if not data:
                        closed.append((sock, chan))
                    else:
                        chan.send(data)

                if chan in r:
                    data = chan.recv(1024)
                    if not data:
                        closed.append((sock, chan))
                    else:
                        sock.send(data)

            for item in set(closed):  # set() will remove duplicates from closed list
                sock, chan = item
                try:
                    sock.close()
                except:
                    pass
                try:
                    chan.close()
                except:
                    pass
                self.notify('INFO', 'Client for %s disconnected' % local_port)
                self._connections.remove(item)

            if closed and not self._connections and time.time() - last_activity > TUNNEL_TIMEOUT:
                self.notify('INFO', 'Closing tunnel to %s:%s for inactivity...' % (self._server[0], self._server[1]) )
                break
        # Time to shutdown:
        for sock, chan in self._connections:
            try:
                sock.close()
            except:
                pass
            try:
                chan.close()
            except:
                pass
        
        self._listen_sock.close()
        self._client.close()
        log_debug("Leaving tunnel thread %s\n" % self.local_port)
        
    def notify(self, msg_type, msg_object):
        log_debug2("tunnel_%i: %s %s\n" % (self.local_port, msg_type, msg_object))
        self.q.put((msg_type, msg_object))
        
    def notify_exception_error(self, msg_type, msg_txt, msg_obj = None):
        self.notify(msg_type, msg_txt)
        log_error("%s\n" % traceback.format_exc())

    def match(self, server, username, target):
        with self.lock:
            return self._server == server and self._username == username and self._target == target


    def _get_ssh_config_path(self):
        paths = []
        user_path = grt.root.wb.options.options['pathtosshconfig'] if grt.root.wb.options.options['pathtosshconfig'] is not None else None
        if user_path:
            paths.append(user_path)
        if platform.system().lower() == "windows":
            paths.append("%s\ssh\config" % mforms.App.get().get_user_data_folder())
            paths.append("%s\ssh\ssh_config" % mforms.App.get().get_user_data_folder())
        else:
            paths.append("~/.ssh/config")
            paths.append("~/.ssh/ssh_config")
            
        for path in paths:
            if os.path.isfile(os.path.expanduser(path)):
                return os.path.expanduser(path)
        else:
            log_debug3("ssh config file not found")
            return None

    def _connect_ssh(self):
        """Create the SSH client and set up the connection.
        
        Any exception coming from paramiko will be notified as an error
        that would cause the failure of the connection. Some of these are:
        
        paramiko.AuthenticationException   --- raised when authentication failed for some reason
        paramiko.PasswordRequiredException --- raised when a password is needed to unlock a private key file;
                                               this is a subclass of paramiko.AuthenticationException
        """
        try:
            
            config = paramiko.config.SSHConfig()
            config_file_path = self._get_ssh_config_path()
            if config_file_path:
                with open(config_file_path) as f:
                    config.parse(f)
                
            opts = config.lookup(self._server[0])
            ssh_known_hosts_file = None
            if "userknownhostsfile" in opts:
                ssh_known_hosts_file = opts["userknownhostsfile"]
            else:
                self._client.get_host_keys().clear()
                ssh_known_hosts_file = '~/.ssh/known_hosts'
                
                if platform.system().lower() == "windows":
                    ssh_known_hosts_file = '%s\ssh\known_hosts' % mforms.App.get().get_user_data_folder()

            try:
                self._client.load_host_keys(os.path.expanduser(ssh_known_hosts_file))
            except IOError, e:
                log_warning("IOError, probably caused by file %s not found, the message was: %s\n" % (ssh_known_hosts_file, e))

            if "stricthostkeychecking" in opts and opts["stricthostkeychecking"].lower() == "no":
                self._client.set_missing_host_key_policy(WarningPolicy())
            else:
                self._client.set_missing_host_key_policy(StoreIfConfirmedPolicy())
                
            has_key = bool(self._keyfile)
            self._client.connect(self._server[0], self._server[1], username=self._username,
                                 key_filename=self._keyfile, password=self._password,
                                 look_for_keys=has_key, allow_agent=has_key)
        except paramiko.BadHostKeyException, exc:
            self.notify_exception_error('ERROR',format_bad_host_exception(exc, '%s\ssh\known_hosts' % mforms.App.get().get_user_data_folder() if platform.system().lower() == "windows" else "~/.ssh/known_hosts file"))
            return False
        except paramiko.BadAuthenticationType, exc:
            self.notify_exception_error('ERROR', "Bad authentication type, the server is not accepting this type of authentication.\nAllowed ones are:\n %s" % exc.allowed_types, sys.exc_info());
            return False
        except paramiko.AuthenticationException, exc:
            self.notify_exception_error('ERROR', "Authentication failed, please check credentials.\nPlease refer to logs for details", sys.exc_info())
            return False
        except socket.gaierror, exc:
            self.notify_exception_error('ERROR', "Error connecting to SSH server: %s\nPlease refer to logs for details." % str(exc))
            return False
        except paramiko.ChannelException, exc:
            self.notify_exception_error('ERROR', "Error connecting SSH channel.\nPlease refer to logs for details: %s" % str(exc), sys.exc_info())
            return False
        except SSHFingerprintNewError, exc:
            self.notify_exception_error('KEY_ERROR', { 'msg': "The authenticity of host '%(0)s (%(0)s)' can't be established.\n%(1)s key fingerprint is %(2)s\nAre you sure you want to continue connecting?"  % {'0': "%s:%s" % (self._server[0], self._server[1]), '1': exc.key.get_name(), '2': exc.fingerprint}, 'obj': exc})
            return False
        except IOError, exc:
            #Io should be report to the user, so maybe he will be able to fix this issue
            self.notify_exception_error('IO_ERROR', "IO Error: %s.\n Please refer to logs for details." % str(exc), sys.exc_info())
            return False

        except Exception, exc:
            self.notify_exception_error('ERROR', "Authentication error, unhandled exception caught in tunnel manager, please refer to logs for details", sys.exc_info())
            return False
        else:
            log_debug("connect_ssh2 OK\n")
            return True

    def close(self):
        self.notify('INFO', 'Closing tunnel')
        self._listen_sock.close()
        self._shutdown = True

    def accept_client(self):
        try:
            local_sock, peeraddr = self._listen_sock.accept()
        except Exception, e:
            self.notify_exception_error('ERROR', 'Error accepting new tunnel client: %r' % e,sys.exc_info())
            return
        self.notify('INFO', 'Client connection established')

        transport = self._client.get_transport()

        try:
            sshchan = transport.open_channel('direct-tcpip', self._target, local_sock.getpeername())
        except paramiko.ChannelException, exc:
            self.notify_exception_error('ERROR', 'Could not open port forwarding SSH channel: %s' % exc)
            local_sock.close()
            return
        except Exception, e:
            self.notify_exception_error('ERROR', 'Remote connection to %s:%d failed: %r' % (self._target[0], self._target[1], e), sys.exc_info())
            local_sock.close()
            return

        if sshchan is None:
            self.notify_exception_error('ERROR', 'Remote connection to %s:%d was rejected by the SSH server.' % (self._target[0], self._target[1]), sys.exc_info())
            local_sock.close()
            return

        self.notify('INFO', 'Tunnel now open %r -> %r -> %r' % (local_sock.getpeername(), sshchan.getpeername(), self._target))

        self._connections.append((local_sock, sshchan))

class TunnelManager:
    def __init__(self):
        self.tunnel_by_port = {}

        self.inpipe = sys.stdin
        self.outpipe = sys.stdout

    def _address_port_tuple(self, raw_address, default_port):
        if type(raw_address) is str:
            if ':' in raw_address:
                address, port = raw_address.split(':', 1)
                try:
                    port = int(port)
                except:
                    port = default_port
                return (address, port)
            else:
                return (raw_address, default_port)
        else:
            return raw_address

    def lookup_tunnel(self, server, username, target):
        server = self._address_port_tuple(server, default_port=SSH_PORT)
        target = self._address_port_tuple(target, default_port=REMOTE_PORT)

        for port, tunnel in self.tunnel_by_port.iteritems():
            if tunnel.match(server, username, target) and tunnel.isAlive():
                with tunnel.lock:
                    return tunnel.local_port
        return None

    def open_tunnel(self, server, username, password, keyfile, target):
        try:
            port = self.open_ssh(server, username, password, keyfile, target)
        except Exception:
            traceback.print_exc()
            return (False, str(traceback.format_exc()))
        return (True, port)
   
    def open_ssh(self, server, username, password, keyfile, target):
        server = self._address_port_tuple(server, default_port=SSH_PORT)
        target = self._address_port_tuple(target, default_port=REMOTE_PORT)
        
        password = password or ''
        keyfile  = keyfile  or None

        if keyfile is not None:
            keyfile = keyfile.decode('utf-8')

        found = None
        for tunnel in self.tunnel_by_port.itervalues():
            if tunnel.match(server, username, target) and tunnel.isAlive():
                found = tunnel
                break

        if found:
            with tunnel.lock:
                log_debug('Reusing tunnel at port %d' % tunnel.local_port)
                return tunnel.local_port
        else:
            tunnel = Tunnel(Queue.Queue(), server, username, target, password, keyfile)
            tunnel.start()
            tunnel.port_is_set.wait()
            with tunnel.lock:
                port = tunnel.local_port
            self.tunnel_by_port[port] = tunnel
            return port


    def wait_connection(self, port):
        tunnel = self.tunnel_by_port.get(port)
        if not tunnel:
            return 'Could not find a tunnel for port %d' % port
        error = None
        tunnel.port_is_set.wait()
        if tunnel.isAlive():
            while True:
                # Process any message in queue. Every retrieved message is printed.
                # If an error is detected in the queue, exit returning its message:
                try:
                    msg_type, msg = tunnel.q.get_nowait()
                except Queue.Empty:
                    continue
                else:
                    if msg_type == 'KEY_ERROR':
                        if mforms.Utilities.show_message("SSH Server Fingerprint Missing", msg['msg'], "Continue", "Cancel", "") == mforms.ResultOk:
                            msg['obj'].client._host_keys.add(msg['obj'].hostname, msg['obj'].key.get_name(), msg['obj'].key)
                            if msg['obj'].client._host_keys_filename is not None:
                                try:
                                    if os.path.isdir(os.path.dirname(msg['obj'].client._host_keys_filename)) == False:
                                        log_warning("Host_keys directory is missing, recreating it\n")
                                        os.makedirs(os.path.dirname(msg['obj'].client._host_keys_filename))
                                    if os.path.exists(msg['obj'].client._host_keys_filename) == False:
                                        log_warning("Host_keys file is missing, recreating it\n")
                                        open(msg['obj'].client._host_keys_filename, 'a').close()
                                    msg['obj'].client.save_host_keys(msg['obj'].client._host_keys_filename)
                                    log_warning("Successfully saved host_keys file.\n")
                                except IOError, e:
                                    error = str(e)
                                    break;
                            error = "Server key has been stored"
                        else:
                            error = "User cancelled"
                        break # Exit returning the error message
                    elif msg_type == 'IO_ERROR':
                        error = msg
                        break # Exit returning the error message
                    else:
                        time.sleep(0.3)
                    _msg = msg
                    if type(msg) is tuple:
                        msg = '\n' + ''.join(traceback.format_exception(*msg))
                        _msg = str(_msg[1])
                    log_debug("%s: %s\n" % (msg_type, msg))
                    if msg_type == 'ERROR':
                        error = _msg
                        break  # Exit returning the error message
                
                if (not tunnel.is_connecting() or not tunnel.isAlive()) and tunnel.q.empty():  
                    break
                time.sleep(0.3)
        log_debug("returning from wait_connection(%s): %s\n" % (port, error))
        return error


    def get_message(self, port):
        if port not in self.tunnel_by_port:
            log_error("Looking up invalid port %s\n" % port)
            return None
        tunnel = self.tunnel_by_port[port]
        try:
            return tunnel.q.get_nowait()
        except Queue.Empty:
            return None


    def set_keepalive(self, port, keepalive):
        if keepalive == 0:
            log_info("SSH KeepAlive setting skipped.\n")
            return
        tunnel = self.tunnel_by_port.get(port)
        if not tunnel:
            log_error("Looking up invalid port %s\n" % port)
            return
        transport = tunnel._client.get_transport()
        if transport is None:
            log_error("SSHTransport not ready yet %d\n" % port)
            return

        transport.set_keepalive(keepalive)

    def close(self, port):
        pass
        # tunnels auto-close when inactive
        #tunnel = self.tunnel_by_port.get(port, None)
        #if tunnel:
        #    tunnel.num_clients -= 1
        #    if tunnel.num_clients == 0:
        #        tunnel.close()
        #        del self.tunnel_by_port[port]

    def send(self, code, arg=''):
        if arg:
            self.outpipe.write(code + ' ' + arg + '\n')
        else:
            self.outpipe.write(code + '\n')
        self.outpipe.flush()
    
    def shutdown(self):
        for tunnel in self.tunnel_by_port.itervalues():
            tunnel.close()
            tunnel.join()

    # FIXME: It seems that this function is never called. Should we remove it?
    def wait_requests(self):
        #print "SSH Tunnel Manager started, waiting for requests..."
        self.send("READY")
        while True:
            request = self.inpipe.readline()
            if not request:
                #print "Exiting tunnel manager..."
                break
            try:
                cmd, args = eval(request, {}, {})
            except:
                self.send("ERROR", "Invalid request")
                continue
            if cmd == "LOOKUP":
                try:
                    port = self.lookup_tunnel(*args)
                    if port is not None:
                        self.send("OK", str(port))
                    else:
                        self.send("ERROR", "not found")
                except Exception, exc:
                    self.send("ERROR", str(exc))
            elif cmd == "OPENSSH":
                try:
                    port = self.open_ssh(*args)
                    self.send("OK", str(port))
                except Exception, exc:
                    self.send("ERROR", str(exc))
            elif cmd == "CLOSE":
                #self.close(args[0])
                self.send("OK")
            elif cmd == "WAIT":
                # wait for the SSH connection to be established
                error = self.wait_connection(args)
                if not error:
                    self.send("OK")
                else:
                    self.send("ERROR "+error)
            elif cmd == "MESSAGE":
                msg = self.get_message(args)
                if msg:
                    self.send(msg)
                else:
                    self.send("NONE")
            else:
                log_error("Invalid request %s\n" % request)
                self.send("ERROR", "Invalid request")

"""
if "--single" in sys.argv:
    target = sys.argv[2]
    if "-pw" in sys.argv:
        password = sys.argv[sys.argv.index("-pw")+1]
    else:
        password = None
    if "-i" in sys.argv:
        keyfile = sys.argv[sys.argv.index("-i")+1]
    else:
        keyfile = None
    server = sys.argv[-1]

    tunnel = Tunnel(None, False)
    if "@" in server:
        username, server = server.split("@", 1)
    else:
        username = ""
    print "Starting tunnel..."
    if type(server) == str:
        if ':' in server:
            server = server.split(":", 1)
            server = (server[0], int(server[1]))
        else:
            server = (server, SSH_PORT)
    if type(target) == str:
        if ':' in target:
            target = target.split(":", 1)
            target = (target[0], int(target[1]))
        else:
            target = (target, REMOTE_PORT)
    
    tunnel.start(server, username, password or "", keyfile, target)

else:
    tm = TunnelManager()
    sys.stdout = sys.stderr
    try:
        tm.wait_requests()
    except KeyboardInterrupt, e:
	    pass

"""



# Copyright (c) 2013, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA


import mforms


def insert_item_to_plugin_context_menu(menu, item):
    bottom = menu.find_item("bottom_plugins_separator")
    if bottom:
        index = menu.get_item_index(bottom)
        if index > 0 and not menu.find_item("top_plugins_separator"):
            sep = mforms.newMenuItem("", mforms.SeparatorMenuItem)
            sep.set_name("top_plugins_separator")
            menu.insert_item(index, sep)
            index += 1
        menu.insert_item(index, item)



variable_groups = [ ('mysqlx_port', []),
  ('mysqlx_max_connections', []),
  ('mysqlx_initial_incoming_queue_bytes', []),
  ('Aborted_clients', ['Networking/Stats']),
  ('Aborted_connects', ['Networking/Stats']),
  ('Audit_log_current_size', ['General']),
  ('Audit_log_event_max_drop_size', ['General']),
  ('Audit_log_events', ['General']),
  ('Audit_log_events_filtered', ['General']),
  ('Audit_log_events_lost', ['General']),
  ('Audit_log_events_written', ['General']),
  ('Audit_log_total_size', ['General']),
  ('Audit_log_write_waits', ['General']),
  ('Binlog_cache_disk_use', ['Binlog']),
  ('Binlog_cache_use', ['Binlog']),
  ('Binlog_stmt_cache_disk_use', ['Binlog']),
  ('Binlog_stmt_cache_use', ['Binlog']),
  ('Bytes_received', ['Networking/Stats']),
  ('Bytes_sent', ['Networking/Stats']),
  ('Com_admin_commands', ['Commands/Admin']),
  ('Com_alter_db', ['Commands/DDL']),
  ('Com_alter_db_upgrade', ['Commands/DDL']),
  ('Com_alter_event', ['Commands/DDL']),
  ('Com_alter_function', ['Commands/DDL']),
  ('Com_alter_procedure', ['Commands/DDL']),
  ('Com_alter_server', ['Commands/DDL']),
  ('Com_alter_table', ['Commands/DDL']),
  ('Com_alter_tablespace', ['Commands/DDL']),
  ('Com_alter_user', ['Commands/Admin']),
  ('Com_analyze', ['Commands/General']),
  ('Com_assign_to_keycache', ['Commands/General']),
  ('Com_backup', []),
  ('Com_backup_table', []),
  ('Com_begin', ['Commands/Transaction']),
  ('Com_binlog', ['Binlog', 'Commands/General']),
  ('Com_call_procedure', ['Commands/DML']),
  ('Com_change_db', ['Commands/General']),
  ('Com_change_master', ['Commands/Admin']),
  ('Com_change_repl_filter', []),
  ('Com_check', ['Commands/Admin']),
  ('Com_checksum', ['Commands/General']),
  ('Com_commit', ['Commands/General']),
  ('Com_create_db', ['Commands/DDL']),
  ('Com_create_event', ['Commands/DDL']),
  ('Com_create_function', ['Commands/DDL']),
  ('Com_create_index', ['Commands/DDL']),
  ('Com_create_procedure', ['Commands/DDL']),
  ('Com_create_server', ['Commands/DDL']),
  ('Com_create_table', ['Commands/DDL']),
  ('Com_create_trigger', ['Commands/DDL']),
  ('Com_create_udf', ['Commands/Admin']),
  ('Com_create_user', ['Commands/Admin']),
  ('Com_create_view', ['Commands/DDL']),
  ('Com_dealloc_sql', ['Commands/DML']),
  ('Com_delete', ['Commands/DML']),
  ('Com_delete_multi', ['Commands/DML']),
  ('Com_do', ['Commands/DML']),
  ('Com_drop_db', ['Commands/DDL']),
  ('Com_drop_event', ['Commands/DDL']),
  ('Com_drop_function', ['Commands/DDL']),
  ('Com_drop_index', ['Commands/DDL']),
  ('Com_drop_procedure', ['Commands/DDL']),
  ('Com_drop_server', ['Commands/DDL']),
  ('Com_drop_table', ['Commands/DDL']),
  ('Com_drop_trigger', ['Commands/DDL']),
  ('Com_drop_user', ['Commands/DDL']),
  ('Com_drop_view', ['Commands/DDL']),
  ('Com_empty_query', ['Commands/General']),
  ('Com_execute_sql', ['Commands/General']),
  ('Com_explain_other', ['Commands/General']),
  ('Com_flush', ['Commands/General']),
  ('Com_get_diagnostics', ['Commands/Admin']),
  ('Com_grant', ['Commands/Admin']),
  ('Com_ha_close', ['Commands/Admin']),
  ('Com_ha_open', ['Commands/Admin']),
  ('Com_ha_read', ['Commands/Admin']),
  ('Com_help', ['Commands/General']),
  ('Com_insert', ['Commands/DML']),
  ('Com_insert_select', ['Commands/DML']),
  ('Com_install_plugin', ['Commands/Admin']),
  ('Com_kill', ['Commands/Admin']),
  ('Com_load', ['Commands/Admin']),
  ('Com_load_master_data', ['Commands/Admin']),
  ('Com_load_master_table', ['Commands/Admin']),
  ('Com_lock_tables', ['Commands/General']),
  ('Com_optimize', ['Commands/Admin']),
  ('Com_preload_keys', ['Commands/Admin']),
  ('Com_prepare_sql', ['Commands/DML']),
  ('Com_purge', ['Commands/Admin']),
  ('Com_purge_before_date', ['Commands/Admin']),
  ('Com_purge_bup_log', ['Commands/Admin']),
  ('Com_release_savepoint', ['Commands/Admin']),
  ('Com_rename_table', ['Commands/DDL']),
  ('Com_rename_user', ['Commands/Admin']),
  ('Com_repair', ['Commands/Admin']),
  ('Com_replace', ['Commands/DML']),
  ('Com_replace_select', ['Commands/DML']),
  ('Com_reset', ['Commands/Admin']),
  ('Com_resignal', ['Commands/Admin']),
  ('Com_restore', ['Commands/Admin']),
  ('Com_restore_table', ['Commands/Admin']),
  ('Com_revoke', ['Commands/Admin']),
  ('Com_revoke_all', ['Commands/Admin']),
  ('Com_rollback', ['Commands/Transaction']),
  ('Com_rollback_to_savepoint', ['Commands/Transaction']),
  ('Com_savepoint', ['Commands/Transaction']),
  ('Com_select', ['Commands/General']),
  ('Com_set_option', ['Commands/General']),
  ('Com_show_authors', ['Commands/Show']),
  ('Com_show_binlog_events', ['Commands/Show']),
  ('Com_show_binlogs', ['Commands/Show']),
  ('Com_show_charsets', ['Commands/Show']),
  ('Com_show_collations', ['Commands/Show']),
  ('Com_show_column_types', ['Commands/Show']),
  ('Com_show_contributors', ['Commands/Show']),
  ('Com_show_create_db', ['Commands/Show']),
  ('Com_show_create_event', ['Commands/Show']),
  ('Com_show_create_func', ['Commands/Show']),
  ('Com_show_create_proc', ['Commands/Show']),
  ('Com_show_create_table', ['Commands/Show']),
  ('Com_show_create_trigger', ['Commands/Show']),
  ('Com_show_create_user', ['Commands/Show']),
  ('Com_show_databases', ['Commands/Show']),
  ('Com_show_engine_logs', ['Commands/Show']),
  ('Com_show_engine_mutex', ['Commands/Show']),
  ('Com_show_engine_status', ['Commands/Show']),
  ('Com_show_errors', ['Commands/Show']),
  ('Com_show_events', ['Commands/Show']),
  ('Com_show_fields', ['Commands/Show']),
  ('Com_show_function_code', ['Commands/Show']),
  ('Com_show_function_status', ['Commands/Show']),
  ('Com_show_grants', ['Commands/Show']),
  ('Com_show_innodb_status', ['Commands/Show']),
  ('Com_show_keys', ['Commands/Show']),
  ('Com_show_logs', ['Commands/Show']),
  ('Com_show_master_status', ['Commands/Show']),
  ('Com_show_ndb_status', ['Commands/Show']),
  ('Com_show_new_master', ['Commands/Show']),
  ('Com_show_open_tables', ['Commands/Show']),
  ('Com_show_plugins', ['Commands/Show']),
  ('Com_show_privileges', ['Commands/Show']),
  ('Com_show_procedure_code', ['Commands/Show']),
  ('Com_show_procedure_status', ['Commands/Show']),
  ('Com_show_processlist', ['Commands/Show']),
  ('Com_show_profile', ['Commands/Show']),
  ('Com_show_profiles', ['Commands/Show']),
  ('Com_show_relaylog_events', ['Commands/Show']),
  ('Com_show_slave_hosts', ['Commands/Show']),
  ('Com_show_slave_status', ['Commands/Show']),
  ('Com_show_slave_status_nonblocking', ['Commands/Show']),
  ('Com_show_status', ['Commands/Show']),
  ('Com_show_storage_engines', ['Commands/Show']),
  ('Com_show_table_status', ['Commands/Show']),
  ('Com_show_tables', ['Commands/Show']),
  ('Com_show_triggers', ['Commands/Show']),
  ('Com_show_variables', ['Commands/Show']),
  ('Com_show_warnings', ['Commands/Show']),
  ('Com_shutdown', ['Commands/Show']),
  ('Com_signal', ['Commands/Admin']),
  ('Com_slave_start', ['Commands/Replication']),
  ('Com_slave_stop', ['Commands/Replication']),
  ('Com_stmt_close', ['Commands/Prepared Statement']),
  ('Com_stmt_execute', ['Commands/Prepared Statement']),
  ('Com_stmt_fetch', ['Commands/Prepared Statement']),
  ('Com_stmt_prepare', ['Commands/Prepared Statement']),
  ('Com_stmt_reprepare', ['Commands/Prepared Statement']),
  ('Com_stmt_reset', ['Commands/Prepared Statement']),
  ('Com_stmt_send_long_data', ['Commands/Prepared Statement']),
  ('Com_truncate', ['Commands/DML']),
  ('Com_uninstall_plugin', ['Commands/Admin']),
  ('Com_unlock_tables', ['Commands/Admin']),
  ('Com_update', ['Commands/DML']),
  ('Com_update_multi', ['Commands/DML']),
  ('Com_xa_commit', ['Commands/Transaction']),
  ('Com_xa_end', ['Commands/Transaction']),
  ('Com_xa_prepare', ['Commands/Transaction']),
  ('Com_xa_recover', ['Commands/Transaction']),
  ('Com_xa_rollback', ['Commands/Transaction']),
  ('Com_xa_start', ['Commands/Transaction']),
  ('Compression', ['General']),
  ('Connection_errors_accept', ['Networking/Errors']),
  ('Connection_errors_internal', ['Networking/Errors']),
  ('Connection_errors_max_connections', ['Networking/Errors']),
  ('Connection_errors_peer_addr', ['Networking/Errors']),
  ('Connection_errors_select', ['Networking/Errors']),
  ('Connection_errors_tcpwrap', ['Networking/Errors']),
  ('Connections', ['Networking/Stats']),
  ('Created_tmp_disk_tables', ['General']),
  ('Created_tmp_files', ['General']),
  ('Created_tmp_tables', ['General']),
  ('Delayed_errors', ['General']),
  ('Delayed_insert_threads', ['General']),
  ('Delayed_writes', ['General']),
  ('Firewall_access_denied', ['Firewall']),
  ('Firewall_access_granted', ['Firewall']),
  ('Firewall_cached_entries', ['Firewall']),
  ('Flush_commands', ['General']),
  ('Handler_commit', ['Handler']),
  ('Handler_delete', ['Handler']),
  ('Handler_discover', ['Handler']),
  ('Handler_external_lock', ['Handler']),
  ('Handler_mrr_init', ['Handler']),
  ('Handler_prepare', ['Handler']),
  ('Handler_read_first', ['Handler']),
  ('Handler_read_key', ['Handler']),
  ('Handler_read_last', ['Handler']),
  ('Handler_read_next', ['Handler']),
  ('Handler_read_prev', ['Handler']),
  ('Handler_read_rnd', ['Handler']),
  ('Handler_read_rnd_next', ['Handler']),
  ('Handler_rollback', ['Handler']),
  ('Handler_savepoint', ['Handler']),
  ('Handler_savepoint_rollback', ['Handler']),
  ('Handler_update', ['Handler']),
  ('Handler_write', ['Handler']),
  ('Innodb_available_undo_logs', ['InnoDB/Stats']),
  ('Innodb_buffer_pool_bytes_data', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_bytes_dirty', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_dump_status', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_load_status', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_data', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_dirty', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_flushed', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_free', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_latched', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_misc', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_pages_total', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_read_ahead', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_read_ahead_evicted', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_read_ahead_rnd', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_read_ahead_seq', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_read_requests', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_reads', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_resize_status', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_wait_free', ['InnoDB/Buffer pool']),
  ('Innodb_buffer_pool_write_requests', ['InnoDB/Buffer pool']),
  ('Innodb_data_fsyncs', ['InnoDB/Data']),
  ('Innodb_data_pending_fsyncs', ['InnoDB/Data']),
  ('Innodb_data_pending_reads', ['InnoDB/Data']),
  ('Innodb_data_pending_writes', ['InnoDB/Data']),
  ('Innodb_data_read', ['InnoDB/Data', 'InnoDB/Stats']),
  ('Innodb_data_reads', ['InnoDB/Data', 'InnoDB/Stats']),
  ('Innodb_data_writes', ['InnoDB/Data', 'InnoDB/Stats']),
  ('Innodb_data_written', ['InnoDB/Data', 'InnoDB/Stats']),
  ('Innodb_dblwr_pages_written', ['InnoDB/General']),
  ('Innodb_dblwr_writes', ['InnoDB/General']),
  ('Innodb_have_atomic_builtins', ['InnoDB/General']),
  ('Innodb_have_sync_atomic', ['InnoDB/General']),
  ('Innodb_heap_enabled', ['InnoDB/General']),
  ('Innodb_log_waits', ['InnoDB/Stats']),
  ('Innodb_log_write_requests', ['InnoDB/Stats']),
  ('Innodb_log_writes', ['InnoDB/Stats']),
  ('Innodb_num_open_files', ['InnoDB/Stats']),
  ('Innodb_os_log_fsyncs', ['InnoDB/Stats']),
  ('Innodb_os_log_pending_fsyncs', ['InnoDB/Stats']),
  ('Innodb_os_log_pending_writes', ['InnoDB/Stats']),
  ('Innodb_os_log_written', ['InnoDB/Stats']),
  ('Innodb_page_size', ['InnoDB/Stats']),
  ('Innodb_pages_created', ['InnoDB/Stats']),
  ('Innodb_pages_read', ['InnoDB/Stats']),
  ('Innodb_pages_written', ['InnoDB/Stats']),
  ('Innodb_row_lock_current_waits', ['InnoDB/Stats']),
  ('Innodb_row_lock_time', ['InnoDB/Stats']),
  ('Innodb_row_lock_time_avg', ['InnoDB/Stats']),
  ('Innodb_row_lock_time_max', ['InnoDB/Stats']),
  ('Innodb_row_lock_waits', ['InnoDB/Stats']),
  ('Innodb_rows_deleted', ['InnoDB/Stats']),
  ('Innodb_rows_inserted', ['InnoDB/Stats']),
  ('Innodb_rows_read', ['InnoDB/Stats']),
  ('Innodb_rows_updated', ['InnoDB/Stats']),
  ('Innodb_truncated_status_writes', ['InnoDB/Stats']),
  ('Innodb_wake_ups', ['InnoDB/Stats']),
  ('Key_blocks_not_flushed', ['Keycache']),
  ('Key_blocks_unused', ['Keycache']),
  ('Key_blocks_used', ['Keycache']),
  ('Key_read_requests', ['Keycache']),
  ('Key_reads', ['Keycache']),
  ('Key_write_requests', ['Keycache']),
  ('Key_writes', ['Keycache']),
  ('Last_query_cost', ['Performance']),
  ('Last_query_partial_plans', ['Performance']),
  ('Locked_connects', ['Networking/Stats']),
  ('Max_execution_time_exceeded', ['Networking/Stats']),
  ('Max_execution_time_set', ['Networking/Stats']),
  ('Max_execution_time_set_failed', ['Networking/Stats']),
  ('Max_statement_time_exceeded', ['Networking/Stats']),
  ('Max_statement_time_set', ['Networking/Stats']),
  ('Max_statement_time_set_failed', ['Networking/Stats']),
  ('Max_used_connections', ['Networking/Stats']),
  ('Max_used_connections_time', ['Networking/Stats']),
  ('mecab_charset', []),
  ('ndb-nodeid', ['NDB']),
  ('Ndb_api_bytes_received_count', ['NDB']),
  ('Ndb_api_bytes_received_count_session', ['NDB']),
  ('Ndb_api_bytes_received_count_slave', ['NDB']),
  ('Ndb_api_bytes_sent_count', ['NDB']),
  ('Ndb_api_bytes_sent_count_session', ['NDB']),
  ('Ndb_api_bytes_sent_count_slave', ['NDB']),
  ('Ndb_api_event_bytes_count', ['NDB']),
  ('Ndb_api_event_bytes_count_injector', ['NDB']),
  ('Ndb_api_event_data_count', ['NDB']),
  ('Ndb_api_event_data_count_injector', ['NDB']),
  ('Ndb_api_event_nondata_count', ['NDB']),
  ('Ndb_api_event_nondata_count_injector', ['NDB']),
  ('Ndb_api_pk_op_count', ['NDB']),
  ('Ndb_api_pk_op_count_session', ['NDB']),
  ('Ndb_api_pk_op_count_slave', ['NDB']),
  ('Ndb_api_pruned_scan_count', ['NDB']),
  ('Ndb_api_pruned_scan_count_session', ['NDB']),
  ('Ndb_api_pruned_scan_count_slave', ['NDB']),
  ('Ndb_api_range_scan_count', ['NDB']),
  ('Ndb_api_range_scan_count_session', ['NDB']),
  ('Ndb_api_range_scan_count_slave', ['NDB']),
  ('Ndb_api_read_row_count', ['NDB']),
  ('Ndb_api_read_row_count_session', ['NDB']),
  ('Ndb_api_read_row_count_slave', ['NDB']),
  ('Ndb_api_scan_batch_count', ['NDB']),
  ('Ndb_api_scan_batch_count_session', ['NDB']),
  ('Ndb_api_scan_batch_count_slave', ['NDB']),
  ('Ndb_api_table_scan_count', ['NDB']),
  ('Ndb_api_table_scan_count_session', ['NDB']),
  ('Ndb_api_table_scan_count_slave', ['NDB']),
  ('Ndb_api_trans_abort_count', ['NDB']),
  ('Ndb_api_trans_abort_count_session', ['NDB']),
  ('Ndb_api_trans_abort_count_slave', ['NDB']),
  ('Ndb_api_trans_close_count', ['NDB']),
  ('Ndb_api_trans_close_count_session', ['NDB']),
  ('Ndb_api_trans_close_count_slave', ['NDB']),
  ('Ndb_api_trans_commit_count', ['NDB']),
  ('Ndb_api_trans_commit_count_session', ['NDB']),
  ('Ndb_api_trans_commit_count_slave', ['NDB']),
  ('Ndb_api_trans_local_read_row_count', ['NDB']),
  ('Ndb_api_trans_local_read_row_count_session', ['NDB']),
  ('Ndb_api_trans_local_read_row_count_slave', ['NDB']),
  ('Ndb_api_trans_start_count', ['NDB']),
  ('Ndb_api_trans_start_count_session', ['NDB']),
  ('Ndb_api_trans_start_count_slave', ['NDB']),
  ('Ndb_api_uk_op_count', ['NDB']),
  ('Ndb_api_uk_op_count_session', ['NDB']),
  ('Ndb_api_uk_op_count_slave', ['NDB']),
  ('Ndb_api_wait_exec_complete_count', ['NDB']),
  ('Ndb_api_wait_exec_complete_count_session', ['NDB']),
  ('Ndb_api_wait_exec_complete_count_slave', ['NDB']),
  ('Ndb_api_wait_meta_request_count', ['NDB']),
  ('Ndb_api_wait_meta_request_count_session', ['NDB']),
  ('Ndb_api_wait_meta_request_count_slave', ['NDB']),
  ('Ndb_api_wait_nanos_count', ['NDB']),
  ('Ndb_api_wait_nanos_count_session', ['NDB']),
  ('Ndb_api_wait_nanos_count_slave', ['NDB']),
  ('Ndb_api_wait_scan_result_count', ['NDB']),
  ('Ndb_api_wait_scan_result_count_session', ['NDB']),
  ('Ndb_api_wait_scan_result_count_slave', ['NDB']),
  ('Ndb_cluster_node_id', ['NDB']),
  ('Ndb_config_from_host', ['NDB']),
  ('Ndb_config_from_port', ['NDB']),
  ('Ndb_conflict_fn_epoch', ['NDB']),
  ('Ndb_conflict_fn_epoch2', ['NDB']),
  ('Ndb_conflict_fn_epoch2_trans', ['NDB']),
  ('Ndb_conflict_fn_epoch_trans', ['NDB']),
  ('Ndb_conflict_fn_max', ['NDB']),
  ('Ndb_conflict_fn_max_del_win', ['NDB']),
  ('Ndb_conflict_fn_old', ['NDB']),
  ('Ndb_conflict_last_stable_epoch', ['NDB']),
  ('Ndb_conflict_reflected_op_discard_count', ['NDB']),
  ('Ndb_conflict_reflected_op_prepare_count', ['NDB']),
  ('Ndb_conflict_refresh_op_count', ['NDB']),
  ('Ndb_conflict_trans_conflict_commit_count', ['NDB']),
  ('Ndb_conflict_trans_detect_iter_count', ['NDB']),
  ('Ndb_conflict_trans_reject_count', ['NDB']),
  ('Ndb_conflict_trans_row_conflict_count', ['NDB']),
  ('Ndb_conflict_trans_row_reject_count', ['NDB']),
  ('Ndb_epoch_delete_delete_count', ['NDB']),
  ('Ndb_execute_count', ['NDB']),
  ('Ndb_last_commit_epoch_server', ['NDB']),
  ('Ndb_last_commit_epoch_session', ['NDB']),
  ('Ndb_number_of_data_nodes', ['NDB']),
  ('Ndb_pruned_scan_count', ['NDB']),
  ('Ndb_pushed_queries_defined', ['NDB']),
  ('Ndb_pushed_queries_dropped', ['NDB']),
  ('Ndb_pushed_queries_executed', ['NDB']),
  ('Ndb_pushed_reads', ['NDB']),
  ('ndb_recv_thread_activation_threshold', ['NDB']),
  ('Ndb_scan_count', ['NDB']),
  ('Not_flushed_delayed_rows', ['General']),
  ('Ongoing_anonymous_gtid_violating_transaction_count', []),
  ('Ongoing_anonymous_transaction_count', []),
  ('Ongoing_automatic_gtid_violating_transaction_count', []),
  ('Open_files', ['General']),
  ('Open_streams', ['General']),
  ('Open_table_definitions', ['General']),
  ('Open_tables', ['General']),
  ('Opened_files', ['General']),
  ('Opened_table_definitions', ['General']),
  ('Opened_tables', ['General']),
  ('Performance_schema_accounts_lost', ['Performance schema']),
  ('Performance_schema_cond_classes_lost', ['Performance schema']),
  ('Performance_schema_cond_instances_lost', ['Performance schema']),
  ('Performance_schema_digest_lost', ['Performance schema']),
  ('Performance_schema_file_classes_lost', ['Performance schema']),
  ('Performance_schema_file_handles_lost', ['Performance schema']),
  ('Performance_schema_file_instances_lost', ['Performance schema']),
  ('Performance_schema_hosts_lost', ['Performance schema']),
  ('Performance_schema_index_stat_lost', ['Performance schema']),
  ('Performance_schema_locker_lost', ['Performance schema']),
  ('Performance_schema_memory_classes_lost', ['Performance schema']),
  ('Performance_schema_metadata_lock_lost', ['Performance schema']),
  ('Performance_schema_mutex_classes_lost', ['Performance schema']),
  ('Performance_schema_mutex_instances_lost', ['Performance schema']),
  ('Performance_schema_nested_statement_lost', ['Performance schema']),
  ('Performance_schema_prepared_statements_lost', ['Performance schema']),
  ('Performance_schema_program_lost', ['Performance schema']),
  ('Performance_schema_rwlock_classes_lost', ['Performance schema']),
  ('Performance_schema_rwlock_instances_lost', ['Performance schema']),
  ('Performance_schema_session_connect_attrs_lost', ['Performance schema']),
  ('Performance_schema_socket_classes_lost', ['Performance schema']),
  ('Performance_schema_socket_instances_lost', ['Performance schema']),
  ('Performance_schema_stage_classes_lost', ['Performance schema']),
  ('Performance_schema_statement_classes_lost', ['Performance schema']),
  ('Performance_schema_table_handles_lost', ['Performance schema']),
  ('Performance_schema_table_instances_lost', ['Performance schema']),
  ('Performance_schema_table_lock_stat_lost', ['Performance schema']),
  ('Performance_schema_thread_classes_lost', ['Performance schema']),
  ('Performance_schema_thread_instances_lost', ['Performance schema']),
  ('Performance_schema_users_lost', ['Performance schema']),
  ('Prepared_stmt_count', ['General']),
  ('Qcache_free_blocks', ['Query cache']),
  ('Qcache_free_memory', ['Query cache']),
  ('Qcache_hits', ['Query cache']),
  ('Qcache_inserts', ['Query cache']),
  ('Qcache_lowmem_prunes', ['Query cache']),
  ('Qcache_not_cached', ['Query cache']),
  ('Qcache_queries_in_cache', ['Query cache']),
  ('Qcache_total_blocks', ['Query cache']),
  ('Queries', ['General']),
  ('Questions', ['General']),
  ('Rewriter_number_loaded_rules', []),
  ('Rewriter_number_reloads', []),
  ('Rewriter_number_rewritten_queries', []),
  ('Rewriter_reload_error', []),
  ('Rpl_semi_sync_master_clients', ['Commands/Replication']),
  ('Rpl_semi_sync_master_net_avg_wait_time', ['Commands/Replication']),
  ('Rpl_semi_sync_master_net_wait_time', ['Commands/Replication']),
  ('Rpl_semi_sync_master_net_waits', ['Commands/Replication']),
  ('Rpl_semi_sync_master_no_times', ['Commands/Replication']),
  ('Rpl_semi_sync_master_no_tx', ['Commands/Replication']),
  ('Rpl_semi_sync_master_status', ['Commands/Replication']),
  ('Rpl_semi_sync_master_timefunc_failures', ['Commands/Replication']),
  ('Rpl_semi_sync_master_tx_avg_wait_time', ['Commands/Replication']),
  ('Rpl_semi_sync_master_tx_wait_time', ['Commands/Replication']),
  ('Rpl_semi_sync_master_tx_waits', ['Commands/Replication']),
  ('Rpl_semi_sync_master_wait_pos_backtraverse', ['Commands/Replication']),
  ('Rpl_semi_sync_master_wait_sessions', ['Commands/Replication']),
  ('Rpl_semi_sync_master_yes_tx', ['Commands/Replication']),
  ('Rpl_semi_sync_slave_status', ['Commands/Replication']),
  ('Rpl_status', ['Commands/Replication']),
  ('Rsa_public_key', []),
  ('Select_full_join', ['General']),
  ('Select_full_range_join', ['General']),
  ('Select_range', ['General']),
  ('Select_range_check', ['General']),
  ('Select_scan', ['General']),
  ('Slave_heartbeat_period', ['Replication']),
  ('Slave_last_heartbeat', ['Replication']),
  ('Slave_open_temp_tables', ['Replication']),
  ('Slave_received_heartbeats', ['Replication']),
  ('Slave_retried_transactions', ['Replication']),
  ('Slave_running', ['Replication']),
  ('Slow_launch_threads', ['Threading']),
  ('Slow_queries', ['General']),
  ('Sort_merge_passes', ['General']),
  ('Sort_range', ['General']),
  ('Sort_rows', ['General']),
  ('Sort_scan', ['General']),
  ('Ssl_accept_renegotiates', ['SSL']),
  ('Ssl_accepts', ['SSL']),
  ('Ssl_callback_cache_hits', ['SSL']),
  ('Ssl_cipher', ['SSL']),
  ('Ssl_cipher_list', ['SSL']),
  ('Ssl_client_connects', ['SSL']),
  ('Ssl_connect_renegotiates', ['SSL']),
  ('Ssl_ctx_verify_depth', ['SSL']),
  ('Ssl_ctx_verify_mode', ['SSL']),
  ('Ssl_default_timeout', ['SSL']),
  ('Ssl_finished_accepts', ['SSL']),
  ('Ssl_finished_connects', ['SSL']),
  ('Ssl_server_not_after', ['SSL']),
  ('Ssl_server_not_before', ['SSL']),
  ('Ssl_session_cache_hits', ['SSL']),
  ('Ssl_session_cache_misses', ['SSL']),
  ('Ssl_session_cache_mode', ['SSL']),
  ('Ssl_session_cache_overflows', ['SSL']),
  ('Ssl_session_cache_size', ['SSL']),
  ('Ssl_session_cache_timeouts', ['SSL']),
  ('Ssl_sessions_reused', ['SSL']),
  ('Ssl_used_session_cache_entries', ['SSL']),
  ('Ssl_verify_depth', ['SSL']),
  ('Ssl_verify_mode', ['SSL']),
  ('Ssl_version', ['SSL']),
  ('Table_locks_immediate', ['General']),
  ('Table_locks_waited', ['General']),
  ('Table_open_cache_hits', ['General']),
  ('Table_open_cache_misses', ['General']),
  ('Table_open_cache_overflows', ['General']),
  ('Tc_log_max_pages_used', ['General']),
  ('Tc_log_page_size', ['General']),
  ('Tc_log_page_waits', ['General']),
  ('Threads_cached', ['Threading']),
  ('Threads_connected', ['Threading']),
  ('Threads_created', ['Threading']),
  ('Threads_running', ['Threading']),
  ('Uptime', ['General']),
  ('Uptime_since_flush_status', ['General']),
  ('validate_password_dictionary_file_last_parsed', []),
  ('validate_password_dictionary_file_words_count', [])]



# Copyright (c) 2012, 2013, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import mforms
from migration_ui_style import newHeaderLabel

class MainView(mforms.Box):
    def __init__(self, main):
        mforms.Box.__init__(self, False)
        self.set_managed()

        self.ui_created = False
        self.main = main

        label = newHeaderLabel("  Project Management")
        self.main.ui_profile.apply_style(label, 'content-label')
        self.add(label, False, False)

        self.main.add_content_page(self, "Projects", "Project Management", "admin_export")

    def page_activated(self, advancing):
        if not self.ui_created:
            self.create_ui()
            self.ui_created = True

    def create_ui(self):
        pass
